local FRACUNIT = FRACUNIT
local TICRATE = TICRATE
local SLIPTIDEHANDLING = 7*FRACUNIT/8
local firstblood = false

local currentVer = 1
local currentRev = 7

local cv_stackfx = CV_RegisterVar({
    name = "rr_stackfx",
    defaultvalue = "On",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR
})

local cv_boostchain = CV_RegisterVar({
    name = "rr_boostchain",
    defaultvalue = "Off",
    possiblevalue = CV_OnOff,
    flags = CV_NETVAR
})

local function leveltimeAnimate(actor, frames, time_per_frame)
    actor.frame = states[actor.state].frame + (leveltime / time_per_frame) % frames
end

freeslot("S_BOOSTSTACKVFX", "SPR_BSSS")

local S_BOOSTSTACKVFX = S_BOOSTSTACKVFX
local vfxdist = 30*FRACUNIT
local vfxminscale = FRACUNIT
local vfxmaxscale = 2*FRACUNIT
local vfxscalediff = vfxmaxscale - vfxminscale

local function getVfxScale(boostcount)
    boostcount = min(boostcount, 4)

    return FixedMul(vfxminscale + FixedMul(vfxscalediff, FixedDiv(boostcount*FRACUNIT, 4*FRACUNIT)), mapobjectscale)
end

states[S_BOOSTSTACKVFX] = {
    sprite = SPR_BSSS,
    frame = A|FF_PAPERSPRITE|FF_TRANS50,
    tics = -1,
}


local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function AddCustomBoost(p, name, s, a, h, stacks)
	if not (p and name) then return end
	if not p.customboosts then return end
	if not p.customboosts[name] then p.customboosts[name] = {} end
	
	local SL = p.customboosts
	
	SL[name].speedboost = s or 0
	SL[name].accelboost = a or 0
	SL[name].handleboost = h or 0
	SL[name].stacks = stacks or 0
end


local function K_ApplyOffroad(p)
	if (p.invincibilitytimer or p.hyudorotimer or p.sneakertimer)
	or (p.growshrinktimer > 0 and SpeedLib.cvars.HugeQuest.value)
		return false
	end
	
	return true
end

local function K_GetKartGameSpeedScalar(t)
	return ((13 + (3*t)) << FRACBITS) / 16
end

local function SL_Spawn(p)
	p.customboosts = {}
	
	p.boostpower = 0
	p.speedboost = 0
	p.accelboost = 0
	p.handleboost = 0
	
	p.startboost = 0
	p.driftboost = 0
	p.sneakertimer = 0
	p.growshrinktimer = 0
	p.invincibilitytimer = 0
	p.hyudorotimer = 0
	p.numboosts = 0
	p.LFsneaker = 0
	p.LFsneakercount = 0
	p.LFfloorsneaker = 0
	p.floorboosts = 0
	p.LFturn = 0
	p.LFdriftinput = 0
	p.LFdrift = 0
	p.LFAIZ = 0
	p.LFrocket = 0
	p.fakeAIZ = 0
	p.speedramp = 0
	--spoofing speedboost here
	p.fakespeedboost = 0
	p.oldAccelInput = 0
end

local function K_MomentumAngleEx(mo, threshold)
	if (FixedHypot(mo.momx, mo.momy) > threshold)
		return R_PointToAngle2(0, 0, mo.momx, mo.momy)
	else
		return mo.angle // default to facing angle, rather than 0
	end
end

local function K_GetKartDriftSparkValue(p)
		return (26*4 + p.kartspeed*2 + (9 - p.kartweight))*8
end

local P_PlayerTouchingSectorSpecial = P_PlayerTouchingSectorSpecial
// sets boostpower, speedboost, accelboost, and handleboost to whatever we need it to be
local function K_GetKartBoostPower(mo)
	if not SpeedLib.running then return end
	if not (mo and mo.valid and mo.player) then return end

	local p = mo.player

	if p.spectator then return end
	
	local SL = SpeedLib.func
	local ks = p.kartstuff
	local SLIPTIDEHANDLING = 7*FRACUNIT/8
	// Light weights have stronger boost stacking -- aka, better metabolism than heavies XD
	local maxmetabolismincrease = FRACUNIT/4
	local metabolism = FRACUNIT + ((p.kartweight-1) * maxmetabolismincrease / 8)

	local boostpower = FRACUNIT
	local speedboost, accelboost, handleboost = 0, 0, 0

	if mapheaderinfo[gamemap] and mapheaderinfo[gamemap].fzero then
		if mapheaderinfo[gamemap].fzerospeed then
			boostpower = mapheaderinfo[gamemap].fzerospeed*FRACUNIT/100
		end

		accelboost = mapheaderinfo[gamemap].fzero/15
	end

	local numboosts = 0

	if ks[k_floorboost] then
		if not p.LFfloorsneaker then
			if P_PlayerTouchingSectorSpecial(p, 2, 6) then  -- for the map specific panels
				p.floorboosts = min($ + 1, 3)
			else
				p.floorboosts = min($ + 1, 1)
			end
		end
		if (p.LFsneaker and ks[k_itemamount] < p.LFsneakercount)
			p.numsneakers = $+1
		end
	elseif ((p.LFsneaker and ks[k_itemamount] < p.LFsneakercount)
	or (p.sneakertimer < ks[k_sneakertimer] and not (p.LFfloorsneaker < ks[k_floorboost]))) then
		p.numsneakers = $ + 1
	elseif not p.sneakertimer then
		p.numsneakers = 0
		p.floorboosts = 0
	end

	if (ks[k_spinouttimer] and ks[k_wipeoutslow] == 1) then // Slow down after you've been bumped
		ks[k_boostpower] = 0
		ks[k_speedboost] = 0
		ks[k_accelboost] = 0
		p.handleboost = 0
		return
	end
	
	// Mini boostpower adjustment because we're negating the speed loss from 
	// size change.
	if p.xismol and p.xismol.smol then
		boostpower = FixedMul(boostpower, FixedDiv(mapobjectscale, p.mo.scale))
	end
	// Offroad is separate, it's difficult to factor it in with a variable value anyway.
	if (SL.Offroad(p) and ks[k_offroad] >= 0) then
		boostpower = FixedDiv(boostpower, FixedMul(ks[k_offroad], SL.GameSpeed(gamespeed)) + FRACUNIT)
	end
	
	if (ks[k_bananadrag] > TICRATE) then
		boostpower = (4*boostpower)/5
	end
	
	//RetroStation: Buckle up, we're about to head into cursed territory
	local function ADDBOOST(s, a, h, st)
		numboosts = $ + 1
		
		s = $ or 0
		a = $ or 0
		h = $ or 0

		if not (s or a or h) then
			numboosts = $ - 1
		end
		
		speedboost = $ + FixedDiv(s, FRACUNIT + (metabolism * max(numboosts - 1, 0)))
		accelboost = $ + FixedDiv(a, FRACUNIT + (metabolism * max(numboosts - 1, 0)))
		handleboost = max($, h)
	end
	
	if (p.sneakertimer) then // Sneaker
		local sneakerpower = FRACUNIT/2
		local hardsneaker = (CV_FindVar("kmp_hardsneakers") and CV_FindVar("kmp_hardsneakers").value)
		if gamespeed == 2 then
			sneakerpower = hardsneaker and $ or 11*FRACUNIT/40
			--print("Sneaker Power: "..sneakerpower)
		end
		for i = 1, (p.numsneakers + p.floorboosts)
			ADDBOOST(sneakerpower, 8*FRACUNIT, SLIPTIDEHANDLING) // + 50% top speed, + 800% acceleration, +50% handling
		end
	end
	
	//Custom boosts here
	for k, v in pairs(p.customboosts or {}) do
		local speedmod = v.speedboost
		local accelmod = v.accelboost
		local handlemod = v.handleboost
		local stacks = v.stacks
		if (speedmod or accelmod or handlemod) then
			if stacks > 1 then
				for i = 1, stacks
					ADDBOOST(speedmod, accelmod, handlemod)
				end
			else
				ADDBOOST(speedmod, accelmod, handlemod)
			end
			--print("Custom Boost applied!")
		end
	end
	
	if (p.invincibilitytimer) then // Invincibility
		// S-Monitor: no extra %
		//RetroStation: Invincibility has WAY too much going for it to also need a sliding scale bonus.
		//Nerfing that shit via CVar.
		local invbase = 3*FRACUNIT/8
		local extra = FRACUNIT / (SpeedLib.cvars.InvNerf.value*350) * (p.invincibilitytimer)
		
		if (p.masteremerald and p.masteremerald.super) then
			-- Stronger base bonus because Super Form, duh :kekw:
			local statTotal = p.kartspeed + p.kartweight

			invbase = Rescale(statTotal*FRACUNIT, 2*FRACUNIT, 18*FRACUNIT, 3*FRACUNIT/5, 2*FRACUNIT/5)
			
			-- More Rings equals more bonus speed
			extra = FRACUNIT/175 * p.ringstuff.numRings
		end

		local invbonus = invbase + extra
			
		ADDBOOST(invbonus, 3*FRACUNIT, SLIPTIDEHANDLING) // + 37.5 + ?% top speed, + 300% acceleration, +25% handling
	end

	if (p.growshrinktimer > 0) then // Grow
		local hugespeed = 0
		local hugeaccel = SpeedLib.cvars.HugeQuest.value and 3*FRACUNIT/2 or 0
		ADDBOOST(hugespeed, hugeaccel, SLIPTIDEHANDLING) // + 0% top speed, + 150% acceleration
	end
	
	if p.spindash and p.spindashboost then
		-- Spindash is sharing driftboost bonus, so just tweak accel here
		ADDBOOST(0, 6*FRACUNIT)
	end

	if (p.startboost) then // Startup Boost
		ADDBOOST(FRACUNIT/2, 4*FRACUNIT)
		--print("Start Boost active")
	end

	if (p.driftboost) then // Drift Boost
		local herbalfolkmedicine = FRACUNIT + FRACUNIT*(p.kartweight-1)/12 + FRACUNIT*(8-p.kartspeed)/32

		local driftSpeed = FRACUNIT/4 // 25% base

		if (p.RRstrongDriftBoost > 0) then
			driftSpeed = FixedMul(driftSpeed, 4*FRACUNIT/3) // 25% . 33%
		end

		driftSpeed = FixedMul(driftSpeed, herbalfolkmedicine)

		ADDBOOST(driftSpeed, 4*FRACUNIT) // + variable top speed, + 400% acceleration
	end
	
	local rs = p.ringstuff
	if rs and rs.boost then // Ring Boost
		ADDBOOST(FRACUNIT/4 + (FRACUNIT / 1750 * (rs.boost)), 4*FRACUNIT) // + 20% + ???% top speed, + 400% acceleration
	end
	
	if (ks[k_eggmanexplode]) then // Ready-to-explode
		ADDBOOST(6*FRACUNIT/20, FRACUNIT) // + 30% top speed, + 100% acceleration,
	end
	
	if p.slipstreamboost then
		local clampedkartspeed = max(min(p.kartspeed, 9), 1)
		local slipspeedboost = Rescale(clampedkartspeed*FRACUNIT, 9*FRACUNIT, FRACUNIT, CV_FindVar("slipstream_minspeedboostpercent").value, CV_FindVar("slipstream_maxspeedboostpercent").value)
		local slipaccelboost = CV_FindVar("slipstream_accelboostpercent").value
		
		speedboost = $ + slipspeedboost
		accelboost = $ + slipaccelboost
		numboosts = $ + 1
	elseif p.slipstreamcharge then
		local clampedkartspeed = max(min(p.kartspeed, 9), 1)
		local draftingboost = Rescale(clampedkartspeed*FRACUNIT, 9*FRACUNIT, FRACUNIT, CV_FindVar("slipstream_mindraftspeedboostpercent").value, CV_FindVar("slipstream_maxdraftspeedboostpercent").value)
		
		speedboost = $ + draftingboost
        numboosts = $ + 1
	end
	
	if p.mo.slopeboost then
		speedboost = $ + p.mo.slopeboost
	end
	
	if p.mo.slopeaccel then
		accelboost = $ + p.mo.slopeaccel
	end

	ks[k_boostpower] = boostpower

	// value smoothing
	-- RetroStation: Outside of handling/accel/drift spark rate, stats don't
	-- factor into much of anything where vanilla's meta is bottom-right.
	-- The meta then shifts towards top-left when factoring in AcroBasics.
	-- Ring Racers had the right idea by making specific parts of the stat grid
	-- excel in one game mechanic or another. I'm taking this a step further
	-- by factoring weight and speed into how well you gain/retain bonus speed.
	-- Lighter accel characters ramp up to top speed easily,
	-- but lose the bonus just as quickly.
	-- Heavier speedy characters, however, take longer to reach bonus top speed,
	-- but retain bonus speed better than their lighter accel counterparts.
	local statTotal = p.kartspeed + p.kartweight
	local rampUp = Rescale(statTotal*FRACUNIT, 2*FRACUNIT, 18*FRACUNIT, 17, 25)
	local decay = Rescale(statTotal*FRACUNIT, 2*FRACUNIT, 18*FRACUNIT, TICRATE/2, 3*TICRATE/2)
	
	if SpeedLib.cvars.SpeedRamp.value == 2
	and not p.speedramp
	and numboosts > 0 then
		p.speedramp = rampUp
	end

	-- RetroStation: Adjust values based on scale; Larger players have more 
	-- inertia than smaller ones, so ramping should be tweaked accordingly.
	rampUp = FixedInt(FixedMul($*FRACUNIT, FixedDiv(p.mo.scale, mapobjectscale)))
	decay = FixedInt(FixedMul($*FRACUNIT, FixedDiv(p.mo.scale, mapobjectscale)))
	
	local speedRamp = (p.fakespeedboost + (speedboost - p.fakespeedboost) / max((rampUp), 1))
	local speedDecay = (p.fakespeedboost + (speedboost - p.fakespeedboost) / max((decay), TICRATE/2))
	
	if not SpeedLib.cvars.SpeedRamp.value then
		speedRamp = (p.fakespeedboost + (speedboost - p.fakespeedboost) / SpeedLib.cvars.RampValue.value)
		speedDecay = (p.fakespeedboost + (speedboost - p.fakespeedboost) / (TICRATE/2))
	elseif SpeedLib.cvars.SpeedRamp.value == 2 then
		speedRamp = (p.fakespeedboost + (speedboost - p.fakespeedboost) / max(p.speedramp, 1))
		speedDecay = (p.fakespeedboost + (speedboost - p.fakespeedboost) / (TICRATE/2))
	end

	if (speedboost > p.fakespeedboost) then
		p.fakespeedboost = speedRamp
	else
		p.fakespeedboost = speedDecay
	end
	
	ks[k_speedboost] = p.fakespeedboost
	ks[k_accelboost] = accelboost
	p.handleboost = handleboost
	p.numboosts = numboosts
	-- print("Boosts active: ".. p.numboosts)
	
	-- Cap speed at 300%; literally no point in going beyond that, even with
	-- low-friction sections
	p.speed = min($, K_GetKartSpeed(p, false)*3)
	
	if cv_boostchain.value then
		if ks[k_sneakertimer] > 1
		and (ks[k_driftboost] > ks[k_sneakertimer]) then
			ks[k_sneakertimer] = ks[k_driftboost]
		elseif ks[k_driftboost] > 0
		and ks[k_driftboost] < ks[k_sneakertimer]
			ks[k_driftboost] = ks[k_sneakertimer]
		end
	end

	p.startboost = ks[k_startboost]
	p.driftboost = ks[k_driftboost]
	p.sneakertimer = ks[k_sneakertimer]
	p.invincibilitytimer = ks[k_invincibilitytimer]
	p.hyudorotimer = ks[k_hyudorotimer]
	p.LFsneaker = (ks[k_itemtype] == 1) and not (ks[k_spinouttimer] or ks[k_wipeoutslow] or ks[k_stolentimer])
	p.LFsneakercount = ks[k_itemamount]
	p.LFfloorsneaker = ks[k_floorboost]
	p.LFrocket = ks[k_rocketsneakertimer]
	
	if HugeQuest
	and ks[k_growshrinktimer] >= 0 then
		p.growshrinktimer = p.hugequest.huge
	else
		p.growshrinktimer = ks[k_growshrinktimer]
	end

	-- Stacked effect
	local stacked = p.numboosts

	if stacked > 1
	and cv_stackfx.value then
		if not (p.boostvfx and p.boostvfx.valid) then
			p.boostvfx = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_THOK)
		end
                
		p.boostvfx.state = S_BOOSTSTACKVFX
		p.boostvfx.color = p.mo.color
		p.boostvfx.destscale = getVfxScale(stacked)
		p.boostvfx.angle = SpeedLib.func.Momentum(p.mo, 6*p.mo.scale) + ANGLE_90
		leveltimeAnimate(p.boostvfx, 5, 3)
                
		local x = p.mo.x + FixedMul(cos(SpeedLib.func.Momentum(p.mo, 6*p.mo.scale)), FixedMul(vfxdist, mapobjectscale))
		local y = p.mo.y + FixedMul(sin(SpeedLib.func.Momentum(p.mo, 6*p.mo.scale)), FixedMul(vfxdist, mapobjectscale))
		local z = p.mo.z
                
		P_MoveOrigin(p.boostvfx, x, y, z)
	elseif p.boostvfx and p.boostvfx.valid then
		p.boostvfx.state = S_INVISIBLE
		p.boostvfx.scale = getVfxScale(max(stacked, 2))
		P_SetOrigin(p.boostvfx, p.mo.x, p.mo.y, p.mo.z)
	end
	
	if p.speedramp
	and p.numboosts then
		p.speedramp = max($ - 1, 1)
	end

	if (p.numboosts > 0) // Boosting after images
		local ghost = P_SpawnGhostMobj(p.mo)
		ghost.extravalue1 = p.numboosts+1
		-- ghost.extravalue2 = (leveltime % ghost.extravalue1)
		ghost.fuse = 3*ghost.extravalue1/2
		ghost.color = p.mo.color
		ghost.frame = $ & ~FF_TRANS50
		ghost.frame = $ | FF_FULLBRIGHT | FF_TRANS20
		ghost.colorized = true
		if (leveltime & 1)
			ghost.flags2 = $ | MF2_DONTDRAW
		else
			ghost.flags2 = $ & ~MF2_DONTDRAW
		end
	elseif p.numboosts == 0 then
		p.speedramp = 0
	end
end

if not SpeedLib then
	print("\x89================================")
	print("\x88* SpeedLib- Kart Edition\x80 by \x89RetroStation")
	print("\x88* Initializing SpeedLib...")
	print("\x88* Library Version \x8F"..currentVer.." (revision "..currentRev..")")
	print("\x89================================")
	
	rawset(_G, "SpeedLib", {
		gLibVersion = currentVer,
		gRevVersion = currentRev,
		func = {},
		cvars = {}
	})
	
	-- Functions
	SpeedLib.func.CustomBoost = AddCustomBoost
	SpeedLib.func.Offroad = K_ApplyOffroad
	SpeedLib.func.GameSpeed = K_GetKartGameSpeedScalar
	SpeedLib.func.BoostPower = K_GetKartBoostPower
	SpeedLib.func.Spawn = SL_Spawn
	SpeedLib.func.Momentum = K_MomentumAngleEx
	SpeedLib.func.Drift = K_GetKartDriftSparkValue

	--CVars--
	SpeedLib.cvars.BoostStack = CV_RegisterVar {
		name = "sl_booststack",
		defaultvalue = "On",
		possiblevalue = CV_OnOff,
		flags = CV_NETVAR,
		description = "Toggles Boost Stacking. Defaults to On."
	}

	SpeedLib.cvars.HugeQuest = CV_RegisterVar {
		name = "sl_hugequest",
		defaultvalue = "On",
		possiblevalue = CV_OnOff,
		flags = CV_NETVAR,
		description = "Grants Grow offroad immunity and speed/accel bonuses."
	}
	
	
	SpeedLib.cvars.InvNerf = CV_RegisterVar {
		name = "sl_extramul",
		defaultvalue = "4",
		possiblevalue = {Min = 4, Max = 10},
		flags = CV_NETVAR,
		description = "Adjusts Invincibility's variable speed bonus; higher multipliers lower the bonus. Defaults to 4."
	}
	
	SpeedLib.cvars.SpeedRamp = CV_RegisterVar {
		name = "sl_rampstyle",
		defaultvalue = "Vanilla",
		possiblevalue = {Vanilla = 0, Modern = 1, Dynamic = 2},
		flags = CV_NETVAR,
		description = "Determines speed value smoothing mode during boosts. Defaults to Vanilla."
	}

	SpeedLib.cvars.RampValue = CV_RegisterVar {
		name = "sl_rampvalue",
		defaultvalue = "1",
		possiblevalue = {MIN = 1, MAX = 17},
		flags = CV_NETVAR,
		description = "Determines speed value smoothing rate. Defaults to 1 (Vanilla)."
	}

	--------
	addHook("ThinkFrame", function()
		if leveltime == 0 then return end
		
		if leveltime == 2 then
			SpeedLib.running = SpeedLib.cvars.BoostStack.value
		end
	
	end)
	
	addHook("NetVars", function(net)
		SpeedLib.running = net($)
	end)

	addHook("MobjThinker", function(mo)
		if not (mo and mo.valid and mo.player) then return end
		if not SpeedLib.running then return end
		SpeedLib.func.BoostPower(mo)
	end, MT_PLAYER)
	

	addHook("PlayerSpawn", function(p)
		SpeedLib.func.Spawn(p)
	end)
end

if SpeedLib.gLibVersion < currentVer
or (SpeedLib.gLibVersion == currentVer and (SpeedLib.gRevVersion ~= nil and SpeedLib.gRevVersion < currentRev)) then
	print("\x89================================")
	print("\x88* SpeedLib\x80 by \x89RetroStation")
	print("\x88* Updating to Library Version \x8F"..currentVer.." (revision "..currentRev..")")
	print("\x89================================")
	SpeedLib.func = {}
	-- Functions
	SpeedLib.func.CustomBoost = AddCustomBoost
	SpeedLib.func.Offroad = K_ApplyOffroad
	SpeedLib.func.GameSpeed = K_GetKartGameSpeedScalar
	SpeedLib.func.BoostPower = K_GetKartBoostPower
	SpeedLib.func.Spawn = SL_Spawn
	
	SpeedLib.gLibVersion = currentVer
	SpeedLib.gRevVersion = currentRev
else
	print("\x89================================")
	print("\x88* SpeedLib is up to date!")
	print("\x89================================")
end